<?php
/* Application (script name): IMegaIpBlock    
 *                   Version: 0.12 
 *                    Author: Igor Mega, http://www.imega.cz/
 *                   Licence: GNU GPL
 *                   
 * Return code 90 = other instance already running
 *              5 = Is not started from command line
 *              4 = Cannot include configuration file
 *            !=0 = error 
 * 
 * Default setting:
 *Install path: /usr/share/imegaipblock
 *  Conf. file: /etc/imegaipblock/imegaipblock.conf.php
 *     PidFile: /var/run/imegaipblock.pid
 *    Log file: /var/log/imegaipblock
 *
 * Info: Script use set_include_path('Conf. file path') !!!    
 * 
 * use: php imegaipblock.php OR php imegaipblock.php -m run  
 *  
 *  swithces: -m [start|clean|stop] 
 *                 start is default, standard script processing
 *                 clean iptables rule only - only when another IMegaIpBlock not running
 *            -i includeConfFilePath
 *            -f confFileName  
 *    
 * Fore rules print use command: iptables -L IMegaIpBlock
 */

/*--- STANDARD sure initialization */
error_reporting(E_ALL | E_STRICT); //display all errors and hints, setting is rewrited in the configuration file
ini_set('display_errors', 'On');   //diplay error, setting is rewrited in the configuration file
echo "IMegaIpBlock script by Igor Mega, http://www.imega.cz\n";
$imblk_logFile = false;
function imblk_report($msg)
{
  global $imblk_conf,$imblk_logFile;
  $msg = '['.date('Y-m-d H:i:s')."] IMegaIpBlock: $msg\n";
  if(!$imblk_logFile)
  {
    if(($imblk_logFile = fopen($imblk_conf['logFile'],"a")) === false)
    {
      echo "ERROR 32 (cannot open log file)";
      return;
    }
  }
  if($imblk_logFile !== false)
  {
    if(fwrite($imblk_logFile,"$msg") === false)
    { 
      echo "  ERROR write to log file";
      $imblk_logFile = false;
      return;
    }
    fflush($imblk_logFile);
  }
}
if(!mb_internal_encoding("UTF-8"))
{
  imblk_report("FATAL ERROR 02 (mb_internal_encoding)");
  exit(2);
}
/*-- 'Don't Panic' setting :) --*/
ini_set('magic_quotes_sybase',"0");//do not use magic quotes
ini_set('magic_quotes_gpc',"0");   //do not use magic quotes for get|post|cookie
set_magic_quotes_runtime(0);       //do not use magic quotes (setting for runtime)
/*--------------------------------*/
/*--- Setting default timezone patch  ---*/
if((function_exists("date_default_timezone_set")) && (function_exists("date_default_timezone_get")))
  @date_default_timezone_set(@date_default_timezone_get());
/*---------------------------------------*/
/*--- parse command line pramaters */
$mode = 'start';//default
if(!isset($include_path))
  $include_path = '/etc/imegaipblock/';
if(!isset($conf_file))  
  $conf_file = 'imegaipblock.conf.php';
if(isset($argv))
{
  $c = count($argv)-1;
  for($i = 1;$i < $c;$i++)
  {
    if($argv[$i] == '-m')
      $mode = $argv[$i+1];
    if($argv[$i] == '-i')
      $include_path = $argv[$i+1];
    if($argv[$i] == '-f')
      $conf_file = $argv[$i+1];
  }
}
else
{
  echo("FATAL ERROR 05 (argv not exists)");
  exit(5);
}
//seting include path 
set_include_path('/etc/imegaipblock/');
if(get_include_path() != '/etc/imegaipblock/')
{
  echo("FATAL ERROR 03 (include folder)");
  exit(3);
}
//include conf file
$include_it = 'allow';//allow use include
$returnCode = require_once 'imegaipblock.conf.php';
if(($returnCode != 7)&&($include_it != 'included'))//include setting file
{
  echo("FATAL ERROR 04 (include configuration file)");
  exit(4);
}
if($imblk_conf['logEveryStartClear'])
{
  copy($imblk_conf['logFile'], $imblk_conf['logFile'].'.old');
  unlink($imblk_conf['logFile']);
}
unset($returnCode,$include_it,$conf_file,$include_path);
/*--- Checking timezone setting ---*/
if((function_exists("date_default_timezone_set")) && (function_exists("date_default_timezone_get")))
{
  if(strlen(date_default_timezone_get()) < 3)
  {
    date_default_timezone_set('Europe/Prague');
    imblk_report("WARNING: Timezone is not set. Will be used 'Europe/Prague'. For change use this code in imegaipblock.conf.php 'date_default_timezone_set('Timezone')'. Info: For list of valid timezones see http://www.php.net/timezones.");
  }
}
/*-----------------------------*/

/*--- Functions ---------------*/
$imblk_terminated = false;
$imblk_exLockFp = false;
function imblk_exclusiveLock()
{
  global $imblk_conf,$imblk_exLockFp;
  $imblk_exLockFp = false;
  if(strlen($imblk_conf['pidFile']) < 3)
  {
    imblk_report("FATAL ERROR 91 (pidFile path too short)");
    exit(91);
  }
  if($imblk_exLockFp = fopen($imblk_conf['pidFile'], "w"))
  {
    if (flock($imblk_exLockFp, LOCK_EX | LOCK_NB ))
    {
      fwrite($imblk_exLockFp, "Started: ".date("Y-d-m H:m:i"));
      return true;
    }
    else
    {
      if(is_file($imblk_conf['pidFile']))
      {
        imblk_report("Error: Another instance already running (Exclusive lock already exists)");
        exit(90);
      }
      else
      {
        imblk_report("FATAL ERROR 92 (cannot get exclusive lock on pid file)");
        exit(92);
      }
    }
  }
  else
  {
    if(is_file($imblk_conf['pidFile']))
    {
      imblk_report("Error: Another instance already running (Cannot open the pid file)");
      exit(90);
    }
    else
    {
      imblk_report("FATAL ERROR 93 (error create pid file, check permission)\n");
      exit(93);
    }
  }
  return false;//?
}
function imblk_releaseLoc()
{
  global $imblk_conf,$imblk_exLockFp;
  if($imblk_exLockFp)
  {
    flock($imblk_exLockFp, LOCK_UN); 
    fclose($imblk_exLockFp);
    $imblk_exLockFp = false;
    unlink($imblk_conf['pidFile']);
  }
}


//private global variables
$imblk_history = Array();
$imblk_blocked = Array();
$imblk_candidate = Array();

function imblk_exec($cmd)
{
  global $imblk_conf;
  $retValue = -1;
  $out = Array();
  if($cmd[0] == '@')
  {
    $cmd = substr($cmd,1);
    if($imblk_conf['logAllCmdCommands'])
      imblk_report(" CMD: $cmd");
    exec($cmd);
  }
  else
  {
    if($imblk_conf['logAllCmdCommands']) 
      imblk_report(" CMD: $cmd");
    exec($cmd,$out,$retValue);
    if($retValue != 0)
    {
      if(count($out) > 0)
        imblk_report("Return: $retValue by command: $cmd\nOutput: " . implode("        \n",$out)."\n");
      else imblk_report("Return: $retValue by command: $cmd\n");
    }
  }
}

function imblk_init()
{
  global $imblk_blocked,$imblk_conf;
  imblk_report("--- Start imblk_init ---");
  $c = count($imblk_conf['rules']);
  for($i=0;$i<$c;$i++)
  {
    if(!isset($imblk_conf['rules'][$i]['logFile']))
    {
      imblk_report("FATAL ERROR 51 (configuration array error)");
      exit(51);
    }
    $imblk_conf['rules'][$i]['offset'] = 2147483647;//2gb - init state 
    $imblk_conf['rules'][$i]['inode'] = 0;
    imblk_report("Log file: {$imblk_conf['rules'][$i]['logFile']}");
    imblk_cmdInit();
  }
  imblk_report("--- End imblk_init ---");
}
function imblk_cmdInit()
{
  global $imblk_conf;
  $cc = count($imblk_conf['setting']['cmdInit']);
  for($ii=0;$ii<$cc;$ii++)
  {
    imblk_exec($imblk_conf['setting']['cmdInit'][$ii]);
  }
}

function imblk_cleanUp()
{
  global $imblk_blocked,$imblk_conf,$imblk_candidate;
  imblk_report("--- Start imblk_cleanUp ---");
  $c = count($imblk_conf['rules']);
  for($i=0;$i<$c;$i++)
  {
    if((!isset($imblk_conf['rules'][$i]['logFile']))||(!isset($imblk_conf['rules'][$i]['params'])))
    {
      imblk_report("FATAL ERROR 51 (configuration array error)");
      exit(51);
    }
    imblk_report("Log file: {$imblk_conf['rules'][$i]['logFile']}");
    imblk_cmdCleanUp();
  }
  imblk_report("--- End imblk_cleanUp ---");
}

function imblk_cmdCleanUp()
{
  global $imblk_conf;
  $cc = count($imblk_conf['setting']['cmdCleanUp']);
  for($ii=0;$ii<$cc;$ii++)
  {
    imblk_exec($imblk_conf['setting']['cmdCleanUp'][$ii]);
  }
}


function imblk_clear()
{
  global $imblk_conf,$imblk_candidate,$imblk_history;
  //clear candidate arrays
  $keys = array_keys($imblk_candidate);
  $c = count($keys);
  $now = time() + $imblk_conf['foundTime'];
  for($i=0;$i<$c;$i++)
  {
    if($imblk_candidate[$keys[$i]][1] > $now)
      unset($imblk_candidate[$keys[$i]]); 
  }
  //clear history
  $keys = array_keys($imblk_history);
  $c = count($keys);
  $now = time();
  for($i=0;$i<$c;$i++)
  {
    if($imblk_history[$keys[$i]][0] < $now)
      unset($imblk_history[$keys[$i]]); 
  }
}

function imblk_unblock()
{
  global $imblk_history,$imblk_blocked,$imblk_conf;
  $keys = array_keys($imblk_blocked);
  $c = count($keys);
  $now = time();
  for($i=0;$i<$c;$i++)
  {
    if($imblk_blocked[$keys[$i]][0] < $now)
    {
      if(array_key_exists($keys[$i],$imblk_history))
      {
        $imblk_history[$keys[$i]][0] = round($now + (($imblk_history[$keys[$i]][0] - $now) * $imblk_conf['repeatBlockHistoryCoef'] + $imblk_conf['historyTime']));
        if($imblk_history[$keys[$i]][0] > ($now + $imblk_conf['maxHistoryTime']))
          $imblk_history[$keys[$i]][0]  = ($now + $imblk_conf['maxHistoryTime']);
      }
      else
      {
        if(count($imblk_history) < $imblk_conf['maxHistoryItems'])
          $imblk_history[$keys[$i]] = Array(time() + $imblk_conf['historyTime'],0);
        else imblk_report("WARNING: History array is full");
      }
      $ipPort = split(':',$keys[$i]);
      if($imblk_blocked[$keys[$i]][1])
        $cmds = $imblk_conf['setting']['cmdUnBlockIpRepeat'];
      else $cmds = $imblk_conf['setting']['cmdUnBlockIp'];
      $c3 = count($cmds); 
      for($i3=0;$i3<$c3;$i3++)
      {
        //commands to block
        $cmd = str_replace('%ip%', $ipPort[0],$cmds[$i3]);
        $cmd = str_replace('%port%', $ipPort[1],$cmd);
        imblk_exec($cmd);
        imblk_report("UNBLOCK PORT: {$ipPort[1]} IP: {$ipPort[0]}");
      }          
      unset($imblk_blocked[$keys[$i]]);
    }   
  }
}

function imblk_block($ip,$port,$ipPort)
{
  global $imblk_blocked,$imblk_conf,$imblk_history;
  if(count($imblk_blocked) < $imblk_conf['maxBlockItems'])
  {
    $blockTime = $imblk_conf['blockTime'];
    $repeat = false;
    if(array_key_exists($ipPort,$imblk_history))
    {
      $t = 1.0+($imblk_history[$ipPort][0] - time()) / $imblk_conf['historyTime'];//ovlivneni rychlosti opakovani
      if($t < 1.00001)//for sure
        $t = 1.0;
      $blockTime = (int) round($imblk_history[$ipPort][1] * $imblk_conf['repeatBlockTimeCoef'] + $imblk_conf['repeatBlockTime']*$t);
      $repeat = true;
      if($blockTime > $imblk_conf['maxBlockTime'])
        $blockTime = $imblk_conf['maxBlockTime'];
      $imblk_history[$ipPort][1] = $blockTime;
    }
    $a = false;
    exec($imblk_conf['setting']['cmdCheck'],$a,$retValue);
    if($retValue != 0)
    {
      imblk_report("WARNING check command return no zero value. CMD: ".$imblk_conf['setting']['cmdCheck']);
      imblk_cmdInit();
    }
    if($repeat)
      $cmds = $imblk_conf['setting']['cmdBlockIpRepeat'];
    else $cmds = $imblk_conf['setting']['cmdBlockIp'];
    $c3 = count($cmds);
    for($i3=0;$i3<$c3;$i3++)
    {
      //commands to block
      $cmd = str_replace('%ip%',$ip,$cmds[$i3]);
      $cmd = str_replace('%port%',$port,$cmd);
      imblk_exec($cmd);
      imblk_report("BLOCK   PORT: $port IP: $ip for $blockTime seconds (".round($blockTime/60)." minutes)");
    }
    $imblk_blocked[$ipPort] = Array(time() + $blockTime,$repeat);
  }else imblk_report("WARNING: Blocked array is full, will not block more ip address");
}

function imblk_getIp($mathes)
{
  global $imblk_conf;
  if(!isset($mathes['host']))
    return false;
  if(in_array($mathes['host'],$imblk_conf['secureIp']))
    return false;
  return $mathes['host'];
}


function imblk_scan()
{
  global $imblk_blocked,$imblk_conf,$imblk_candidate;
  imblk_clear();
  imblk_unblock();
  
  $c = count($imblk_conf['rules']);
  $c3 = count($imblk_conf['setting']['cmdBlockIp']);
  clearstatcache();
  for($i=0;$i<$c;$i++)
  {
    if(($fp = fopen($imblk_conf['rules'][$i]['logFile'], 'r')) === false)
    {
      imblk_report("Cannot open log file: {$imblk_conf['rules']['logFile']}");
      return;
    }
    if(filesize($imblk_conf['rules'][$i]['logFile']) < $imblk_conf['rules'][$i]['offset'])
    {
      $imblk_conf['rules'][$i]['offset'] =  filesize($imblk_conf['rules'][$i]['logFile']);
      $imblk_conf['rules'][$i]['inode']  = fileinode($imblk_conf['rules'][$i]['logFile']);
      imblk_report("Set offset to the end of file: " . $imblk_conf['rules'][$i]['logFile']);
    }
    else
    {
      if($fnode = fileinode($imblk_conf['rules'][$i]['logFile']) != $imblk_conf['rules'][$i]['inode'])
      {
        $imblk_conf['rules'][$i]['offset'] = 0;
        $imblk_conf['rules'][$i]['inode']  = $fnode;
      }
    }

    if(fseek($fp,$imblk_conf['rules'][$i]['offset'],SEEK_SET) != 0)
    {
      imblk_report("Cannot fseek log file: {$imblk_conf['rules'][$i]['logFile']}");
      fclose($fp);
      return;
    }
    while(!feof($fp))
    {
      if(($line = fgets($fp,65536)) === false) 
      {
        if(feof($fp))
          break;
        imblk_report("Cannot fgets log file: {$imblk_conf['rules'][$i]['logFile']}");
        fclose($fp);
        return;
      }
      $params = $imblk_conf['rules'][$i]['params'];
      $cc = count($params);
      for($ii=0;$ii<$cc;$ii++)
      {
        $ccc = count($params[$ii]['regexp']);
        for($iii=0;$iii<$ccc;$iii++)
        {
          $mathes = false;
          if(($ret = preg_match($params[$ii]['regexp'][$iii], $line, $mathes)) !== false )
          {
            if($ret > 0)
            {
              //found
              if(($ip = imblk_getIp($mathes)) === false)
                continue;
              $ipPort = "$ip:". $params[$ii]['port'];
              if(array_key_exists($ipPort,$imblk_blocked))
              {
                //already blocked
                continue;
              }  
              if(array_key_exists($ipPort,$imblk_candidate))
              {
                //already exist any records
                $tick = $imblk_candidate[$ipPort][0];
                if($tick < $imblk_conf['foundToBlock'])
                {
                  //increase candidate counter
                  $imblk_candidate[$ipPort][0] = $tick + 1;
                  $imblk_candidate[$ipPort][1] = time();
                }
                else
                {
                  //block ip:port
                  imblk_block($ip,$params[$ii]['port'],$ipPort);               
                  unset($imblk_candidate[$ipPort]); 
                }
              }
              else
              {
                //new occurs in log
                if(count($imblk_candidate) < $imblk_conf['maxCandidateItems'])
                  $imblk_candidate[$ipPort] = Array(1,time());
                else imblk_report("WARNING: Candidate array is full, will not block (control) more ip address");    
              }
              break;
            }
          }
          else        
          {
            //regular expresion error
            imblk_report("ERROR regexp: {$params[$ii]['regexp'][$iii]}");
          }
        }
      }
    }     
    $imblk_conf['rules'][$i]['offset'] = filesize($imblk_conf['rules'][$i]['logFile']);
    fclose($fp);
  }
}


/*-----------------------------*/
imblk_report("Started mode: $mode");
$imblk_conf['frequence']       = abs((int) $imblk_conf['frequence']);
$imblk_conf['maxBlockItems']   = abs((int) $imblk_conf['maxBlockItems']); 
$imblk_conf['maxHistoryItems'] = abs((int) $imblk_conf['maxHistoryItems']);
$imblk_conf['historyTime']     = abs((int) $imblk_conf['historyTime']);
$imblk_conf['maxBlockTime']    = abs((int) $imblk_conf['maxBlockTime']);
$imblk_conf['repeatBlockTime'] = abs((int) $imblk_conf['repeatBlockTime']);
$imblk_conf['timeToDropAllBlock']   = abs((int) $imblk_conf['timeToDropAllBlock']);
switch($mode)

{
  case 'start':if(!imblk_exclusiveLock())
	       {
                 imblk_report("FATAL ERROR 88 (???)");
                 exit(88);
               }
                
               $start = time() + $imblk_conf['timeToDropAllBlock'];
               imblk_init();
               while(!$imblk_terminated)
               {        
                 if(time() > $start)
                 { //start AND restart, probably synchronize iptables with this script
                   imblk_report("Unblock all at this moment and restart blocking (timeToDropAllBlock is reached)");
                   imblk_cmdInit();
                   $imblk_blocked = Array();
                   $start = time() + $imblk_conf['timeToDropAllBlock'];
                 } 
                 imblk_scan();
                 sleep($imblk_conf['frequence']);
               } 
               if($fpid !== NULL)//?
               {
                 fwrite($fpid, " Finished: " . Date('Y-m-d H:i:s'));
                 fclose($fpid);
               }
               imblk_cleanUp();
               imblk_releaseLoc();
               imblk_report("FINISHED");
               if($imblk_logFile !== false)
                 fclose($imblk_logFile);
               echo "Warning: IMegaIpBlock is not running --- Stopped";
               exit(0);
               break;//?
  case 'clean':if(!imblk_exclusiveLock())
               {
                 imblk_report("FATAL ERROR 88 (???)");
                 exit(88);
               }
               imblk_cleanUp();
               imblk_releaseLoc();
               imblk_report("Clean: FINISHED");
               if($imblk_logFile !== false)
                 fclose($imblk_logFile);
               exit(0);
               break;//?    
  default    : imblk_report("FATAL ERROR 06 (invalid mode)");
               if($imblk_logFile !== false)
                 fclose($imblk_logFile);
               exit(6);
               break;
}
imblk_report("[IMegaIpBlock] FATAL ERROR 99 (no correct end of script)");
exit(99);
?>
